# **References**
# - https://wanasit.github.io/attention-based-sequence-to-sequence-in-keras.html
# - https://github.com/keras-team/keras/blob/master/examples/cnn_seq2seq.py
# Note:
# Scripts that were used to predict SMILES strings described in
# "Virtual Screening for Reactive Natural Products and Their Probable Artifacts of Solvolysis and Oxidation" Tingjun Xu et al.
from __future__ import print_function
from keras.models import Model, load_model
from keras.layers import Input, Convolution1D, Dot, Dense, Activation, Concatenate
import numpy as np
import pymssql
from rdkit import Chem
Model_file_name="model.h5" # File name of the used model.
batch_size = 64  # Batch size for training.
epochs = 500  # Number of epochs to train for.
latent_dim = 256  # Latent dimensionality of the encoding space.
latent_dim_out = 64  # Latent dimensionality of the output encoding space.
# Vectorize the data.
input_texts = []
target_texts = []
input_characters = set()
target_characters = set()
conn = pymssql.connect(server="DatabaseServer", database="TrainDataSet")
cursor = conn.cursor()
cursor.execute('SELECT Reactant, Product from TrainData')  # Read the train data set from sql server database.
row = cursor.fetchone()
while row:
    input_text = row[0]  # Reactant.
    target_text = row[1]  # Product.
    # input_text = '\t' + input_text + ' ' + '\n'
    # target_text = '\t' + target_text + ' ' + '\n'
    input_text = input_text + ' '
    target_text = '\t' + target_text + ' ' + '\n'
    input_texts.append(input_text)
    target_texts.append(target_text)
    for char in input_text:
        if char not in input_characters:
            input_characters.add(char)
    for char in target_text:
        if char not in target_characters:
            target_characters.add(char)
    row = cursor.fetchone()
# conn.close()
input_characters = sorted(list(input_characters))
target_characters = sorted(list(target_characters))
num_encoder_tokens = len(input_characters)
num_decoder_tokens = len(target_characters)
max_encoder_seq_length = max([len(txt) for txt in input_texts])
max_decoder_seq_length = max([len(txt) for txt in target_texts])

input_token_index = dict(
    [(char, i) for i, char in enumerate(input_characters)])
target_token_index = dict(
    [(char, i) for i, char in enumerate(target_characters)])
encoder_input_data = np.zeros(
    (len(input_texts), max_encoder_seq_length, num_encoder_tokens),
    dtype='float32')
for i, input_text in enumerate(input_texts):
    for t, char in enumerate(input_text):
        encoder_input_data[i, t, input_token_index[char]] = 1.
    encoder_input_data[i, t + 1:, input_token_index[' ']] = 1.

model = load_model(Model_file_name)
# Define an input sequence and process it.
encoder_inputs = model.input[0]
encoder_outputs = model.get_layer('conv1d_3').output
encoder_model = Model(encoder_inputs)
# Encoder
decoder_inputs = model.input[1]
decoder_outputs = model.get_layer('conv1d_6').output
marge1 = [encoder_outputs, decoder_outputs]
# Decoder
# Attention
inputdot = model.get_layer('dot_1')
attention = inputdot(marge1)
outdot = model.get_layer('dot_2')
marge2 = [model.get_layer('activation_1').output, encoder_outputs]
attentionout = outdot(marge2)
concatenate = model.get_layer('concatenate_1')
marge3 = [model.get_layer('dot_2').get_output_at(0), decoder_outputs]
concatenateout = concatenate(marge3)
# Output
decoder_dense = model.get_layer('dense_1')
decoder_outputs = decoder_dense(model.get_layer('conv1d_8').output)
# Test
test_smiles = []
test_IDs = []
cursor.execute('SELECT ID, Reactant FROM testdata WHERE len([CanonSmiles])<=100') # Read the test data set from sql server database.
row = cursor.fetchone()
while row:
    ReactantInput = row[1]
    ReactantInput = ReactantInput + ' '
    falg = True
    for t, char in enumerate(ReactantInput):
        try:
            test = input_token_index[char]
        except:
            falg = False
    if falg:
        test_smiles.append(ReactantInput)
        test_IDs.append(row[0])
    row = cursor.fetchone()
encoder_input_data_test = np.zeros(
    (len(test_smiles), max_encoder_seq_length, num_encoder_tokens),
    dtype='float32')
for i, input_text in enumerate(test_smiles):
    for t, char in enumerate(input_text):
        encoder_input_data_test[i, t, input_token_index[char]] = 1.
    encoder_input_data_test[i, t + 1:, input_token_index[' ']] = 1.

reverse_input_char_index = dict(
    (i, char) for char, i in input_token_index.items())
reverse_target_char_index = dict(
    (i, char) for char, i in target_token_index.items())

nb_examples = len(test_smiles)
print('Number of samples:', nb_examples)
in_encoder = encoder_input_data_test
in_decoder = np.zeros(
    (len(test_smiles), max_decoder_seq_length, num_decoder_tokens),
    dtype='float32')

in_decoder[:, 0, target_token_index["\t"]] = 1.

predict = np.zeros(
    (len(test_smiles), max_decoder_seq_length),
    dtype='float32')

for i in range(max_decoder_seq_length - 1):
    predict = model.predict([in_encoder, in_decoder])
    predict = predict.argmax(axis=-1)
    predict_ = predict[:, i].ravel().tolist()
    for j, x in enumerate(predict_):
        in_decoder[j, i + 1, x] = 1
Num = 1
for seq_index in range(nb_examples):
    output_seq = predict[seq_index, :].ravel().tolist()
    decoded = []
    for x in output_seq:
        if reverse_target_char_index[x] == "\n":
            break
        else:
            decoded.append(reverse_target_char_index[x])
    decoded_sentence = "".join(decoded)
    try:
        mol = Chem.MolFromSmiles(decoded_sentence[:-1]) # # Validity of predicted SMILES strings for molecular structure.
        Inchikey = Chem.MolToInchiKey(m) # Chemical structure identification of InchiKey.
    except:
        Inchikey = ''
conn.close()
